import com.cycling74.max.*;
import java.util.Random;

public class TheCreeperScratch3 extends MaxObject {

    //can write a list of up to 100 pitch values
    private static final int trainingLen = 400;
    private int T;
    private int observedLength;
    
    //how random do you want the noise to be?
    private double noiseProb = .15;
    private double randProb;
    private int counter;
    
    //random number generator
    private static Random generator = new Random();
    private static final int minRando = 3000;
    private static final int maxRando = 3100;
    
    //number of states
    private int N = 2;
    
    //number of observations
    private int M = 4;
    
    //other variables
    private static final int maxIters = 500;
    private static double oldLogProb = -1999999999;
    private int listLen;
    private double logProb;
    private int iters = 0;
    private double denom;
    private double numer;
    private boolean running;
    private double probOfSeq;
    private double probOfSeq2;
    
    //declared arrays
    private double[][] a;
    private double[][] b;
    private double[] c;
    private double[][] alpha;
    private double[][] beta;
    private double[] pi;
    private double[][] gamma;
    private double[][][] digamma;
    private double[] aSum;
    private double[] bSum;
    private int[] notes = new int[T]; //This is the LONG training sequence
    private int[] notes2; //This should be for the testing sequences...
    private int[] trainer; //These are for the incoming training sequences...
    private double rowSum;
    private double piSum;
    private double alphaSum;

    public TheCreeperScratch3(int[] maxArgs) {
        declareIO(2,2);
        if (maxArgs.length > 1) {
            N = maxArgs[0];
            M = maxArgs[1];
            System.out.println("N = " + N + " and M = " + M);
        }
        else if (maxArgs.length == 1) {
            N = maxArgs[0];
            System.out.println("N = " + N + " and M = " + M);
        }
        else {
            System.out.println("N = " + N + " and M = " + M);
        }
    }
    
    public void bang() {
        
        a = new double[N][N];
        b = new double[N][M];
        pi = new double[N];
        aSum = new double[N];
        bSum = new double[N];
        
        if (getInlet() == 0) {
            System.out.println("This is the left inlet");
            System.out.println("N = " + N + " and M = " + M);
            
            //a bang resets the patch, set iters=0, sets new probabilities
            iters = 0;
            oldLogProb = -2000;
            
            //define initial pi's
            //System.out.println(" ");
            //System.out.println("These are the pi[i] values, pi[0], pi[1]...");
            piSum = 0;
            for (int i=0; i<N; ++i) {
                pi[i] = generator.nextInt(maxRando-minRando) + minRando;
                piSum = piSum + pi[i];
                //System.out.println(piSum);
            }
        
            System.out.println(" ");
            System.out.println("These are the pi[i] probs, pi[0], pi[1]...");
            for (int i=0; i<N; ++i) {
                pi[i] = pi[i]/piSum;
                System.out.println("pi[" + i + "] is " + pi[i]);
            }
        
            //define initial a's
            //System.out.println(" ");
            //System.out.println("These are the a[i][j] values, a[0][0], a[0][1]...");
            for (int i=0; i<N; ++i) {
                aSum[i] = 0;
                for (int j=0; j<N; ++j) {
                    a[i][j] = generator.nextInt(maxRando-minRando) + minRando;
                    aSum[i] = aSum[i] + a[i][j];
                    //System.out.println(aSum[i]);
                }
            }
        
            System.out.println(" ");
            System.out.println("These are the a[i][j] probs, a[0][0], a[0][1]...");
            for (int i=0; i<N; ++i) {
                for (int j=0; j<N; ++j) {
                    a[i][j] = a[i][j]/aSum[i];
                    System.out.println("a[" + i + "][" + j + "] is " + a[i][j]);
                }
            }
        
            //define initial b's
            //System.out.println(" ");;
            //System.out.println("These are the b[i][k] values, b[0][0], b[0][1]...");
            for (int i=0; i<N; ++i) {
                bSum[i] = 0;
                for (int k=0; k<M; ++k) {
                    b[i][k] = generator.nextInt(maxRando-minRando) + minRando;
                    bSum[i] = bSum[i] + b[i][k];
                    //System.out.println(bSum[i]);
                }
            }
        
            System.out.println(" ");
            System.out.println("These are the b[i][k] probs, b[0][0], b[0][1]...");
            for (int i=0; i<N; ++i) {
                for (int k=0; k<M; ++k) {
                    b[i][k] = b[i][k]/bSum[i];
                    System.out.println("b[" + i + "][" + k + "] is " + b[i][k]);
                }
            }
        
            System.out.println(" ");
            System.out.println("This is the b[0][k] sum, b[0][0]+b[0][1]...");
            rowSum = 0;
            for (int k=0; k<M; ++k) {
                rowSum = rowSum + b[0][k];
            }
            System.out.println(rowSum);

        }
        else if (getInlet() == 1) {
            System.out.println("This is the right inlet");
        }
        
    }
    
    public void list(Atom[] args) {
        
        if (getInlet() == 0) {
        
            listLen = args.length;
            trainer = new int[listLen];
            T = trainingLen;
            notes = new int[T];
            c = new double[T];
            alpha = new double[T][N];
            beta = new double[T][N];
            gamma = new double[T][N];
            digamma = new double[T][N][N];
            running = true;
            iters = 0;
            outlet(0, args);
            trainer = Atom.toInt(args);
            counter = 0;
            
            System.out.println("The length of the incoming data is " + T + ".");
            
            //Tell the user the incoming notes
            for (int i=0; i<listLen; ++i) {
                System.out.println("Note " + i + " is " + trainer[i] + ". ");
            }
            
            //Make a T-length array where the incoming notes are repeated and notes
            //are randomly replaced
            for (int i=0; i<T; ++i) {
                randProb = generator.nextDouble();
                if (randProb <= noiseProb/2) {
                    notes[i] = generator.nextInt(M);
                    //System.out.println("Added"); 
                }
                else if (randProb <= noiseProb && randProb >= noiseProb/2) {
                    notes[i] = generator.nextInt(M);
                    //System.out.println("Replaced");
                    ++counter;
                }
                else {
                    notes[i] = trainer[counter%listLen];
                    //System.out.println("Regular");
                    ++counter;
                }
            }
            
            //Print the last 15 notes of the sequence
            for (int i=T-15; i<T; ++i) {
                System.out.println("The " + i + "th note is " + notes[i]);
            }

            System.out.println("The length of the training data is " + T + ".");

            while (running) {
                //compute alpha[0][i]
                //System.out.println(" ");
                //System.out.println("These are UNscaled alpha[0][i] probs, alpha[0][0], alpha[0][1]... Remember 0 refers to time 0...");
                c[0] = 0;
                for (int i=0; i<N; ++i) {
                    alpha[0][i] = pi[i]*b[i][notes[0]];
                    c[0] = c[0]+alpha[0][i];
                    //System.out.println("alpha[0][" + i + "] is " + alpha[0][i]);
                }
        
                //scale the alpha[0][i]
                //System.out.println(" ");
                //System.out.println("These are scaled alpha[0][i] values, alpha[0][0], alpha[0][1]... Remember 0 refers to time 0...");
                c[0] = 1/c[0];
                for (int i=0; i<N; ++i) {
                    alpha[0][i] = c[0] * alpha[0][i];
                    //System.out.println("alpha[0][" + i + "] is " + alpha[0][i]);
                }
            
                //compute alpha[t][i]
                //System.out.println(" ");
                //System.out.println("These are scaled alpha[t][i] values, alpha[1][0], alpha[1][1]...");
                for (int t=1; t<T; ++t) {
                    c[t] = 0;
                    for (int i=0; i<N; ++i) {
                        alpha[t][i] = 0;
                        for (int j=0; j<N; ++j) {
                            alpha[t][i] = alpha[t][i]+alpha[t-1][j]*a[j][i];    //is this a[j][i] in the right order?? Check it!
                        }
                        alpha[t][i] = alpha[t][i]*b[i][notes[t]];   //Again needs to be an observation
                        c[t] = c[t]+alpha[t][i];
                        //System.out.println("alpha[" + t + "][" + i + "] is " + alpha[t][i]);  //these are the unscaled values of alpha
                    }
                
                    //scale alpha[t][i]
                    c[t] = 1/c[t];
                    for (int i=0; i<N; ++i) {
                        alpha[t][i] = c[t]*alpha[t][i];;
                        //System.out.println("alpha[" + t + "][" + i + "] is " + alpha[t][i]);    //these are the scaled values of alpha
                    }
                }
            
                //Let beta[T-1][i]=1 scaled by c[T-1]
                //System.out.println(" ");
                //System.out.println("These scaled beta[T-1][i] = c[T-1]");
                for (int i=0; i<N; ++i) {
                    //beta[T-1][i]=1; //This is used to see beta probs instead of unscaled values
                    beta[T-1][i]=c[T-1];
                    //System.out.println("beta[" + (T-1) + "][" + i + "] is " + beta[T-1][i]);
                }
            
                //Beta-pass
                //System.out.println(" ");
                //System.out.println("These are scaled beta[t][i] values, beta[T-2][0], beta[T-2][1]...");
                for (int t=T-2; t>=0; --t) {
                    for (int i=0; i<N; ++i) {
                        beta[t][i] = 0;
                        for (int j=0; j<N; ++j) {
                            beta[t][i] = beta[t][i]+a[i][j]*b[j][notes[t+1]]*beta[t+1][j];
                        }
                        //scale beta[t][i] with the same scale factor as alpha[t][i]
                        beta[t][i] = c[t]*beta[t][i]; //Comment this out if want to see unscaled values
                        //System.out.println("beta[" + t + "][" + i + "] is " + beta[t][i]);
                    }
                }
            
                //Compute digamma[t][i][j] and gamma[t][i] note, no scaling is done in these digamma
                //all scaling is done in the alpha and beta passes.
                //System.out.println(" ");
                //System.out.println("These are the values for digamma[t][i][j] and gamma[t][i]");
                for (int t=0; t<T-1; ++t) {
                    denom = 0;
                    for (int i=0; i<N; ++i) {
                        for (int j=0; j<N; ++j) {
                            denom = denom + alpha[t][i]*a[i][j]*b[j][notes[t+1]]*beta[t+1][j];
                        }
                    }
                
                    for (int i=0; i<N; ++i) {
                        gamma[t][i] = 0;
                        for (int j=0; j<N; ++j) {
                            digamma[t][i][j] = (alpha[t][i]*a[i][j]*b[j][notes[t+1]]*beta[t+1][j])/denom;
                            gamma[t][i] = gamma[t][i] + digamma[t][i][j];
                            //System.out.println("digamma[" + t + "][" + i + "][" + j +"] is " + digamma[t][i][j]);
                        }
                        //System.out.println("gamma[" + t + "][" + i + "] is " + gamma[t][i]);
                        //System.out.println(" ");
                    }
                
                    //System.out.println("The denominator P(O|Model) is " + denom);
                }
            
                //Re-estimate pi
                //System.out.println(" ");
                //System.out.println("These are the new values for pi[i]");
                for (int i=0; i<N; ++i) {
                    pi[i] = gamma[0][i];
                    //System.out.println("pi[" + i + "] is " + pi[i]);
                }
            
                //Re-estimate A
                //System.out.println(" ");
                //System.out.println("These are the new values for a[i][j]");
                for (int i=0; i<N; ++i) {
                    for (int j=0; j<N; ++j) {
                        numer = 0;
                        denom = 0;
                        for (int t=0; t<T-1; ++t) {
                            numer = numer + digamma[t][i][j];
                            denom = denom + gamma[t][i];
                        }
                        a[i][j] = numer/denom;
                        //System.out.println("a[" + i + "][" + j + "] is " + a[i][j]);
                    }
                }
            
                //Re-estimate B
                //System.out.println(" ");
                //System.out.println("These are the new values for b[i][k]");
                for (int i=0; i<N; ++i) {
                    for (int k=0; k<M; ++k) {
                        numer = 0;
                        denom = 0;
                        for (int t=0; t<T-1; ++t) {
                            if (notes[t]==k) {
                                numer = numer + gamma[t][i];
                            }
                            denom = denom + gamma[t][i];
                        }
                        b[i][k] = numer/denom;
                        //System.out.println("b[" + i + "][" + k + "] is " + b[i][k]);
                    }
                }
            
                //Compute log[P(O|Model)]
                //System.out.println(" ");
                //System.out.println("This is the new value for log[P(O|Model)]");
                logProb = 0;
                for (int t=0; t<T; ++t) {
                    logProb = logProb + Math.log(c[t]);
                }
                logProb = -logProb;
                //System.out.println("logProb is " + logProb);
                
                //System.out.println(" ");
                //System.out.println("This is the new value for P(O|Model)");
                probOfSeq = Math.exp(logProb);
                //System.out.println("P(O|Model) is " + probOfSeq);
            
                //Choose whether or not to iterate
                iters = iters + 1;
                if (iters<maxIters && logProb>oldLogProb) {
                    oldLogProb = logProb;
                }
                else {
                    running = false;
                }
                
            }
            //System.out.println(" ");    
            //System.out.println("A list in the left inlet");
            System.out.println(" ");
            System.out.println("These are the pi[i] probs, pi[0], pi[1]...");
            for (int i=0; i<N; ++i) {
                System.out.println("pi[" + i + "] is " + pi[i]);
            }
            
            System.out.println(" ");
            System.out.println("These are the a[i][j] probs, a[0][0], a[0][1]...");
            for (int i=0; i<N; ++i) {
                for (int j=0; j<N; ++j) {
                    System.out.println("a[" + i + "][" + j + "] is " + a[i][j]);
                }
            }
        
            System.out.println(" ");
            System.out.println("These are the b[i][k] probs, b[0][0], b[0][1]...");
            for (int i=0; i<N; ++i) {
                for (int k=0; k<M; ++k) {
                    System.out.println("b[" + i + "][" + k + "] is " + b[i][k]);
                }
            }
            
            System.out.println(" ");
            System.out.println("The Creeper has iterated " + iters + " times.");
            
        }
        
        else if (getInlet() == 1) {
            
            T = args.length;
            System.out.println("The length of the incoming data is " + T + ".");
            notes = new int[T];
            c = new double[T];
            alpha = new double[T][N];
            beta = new double[T][N];
            gamma = new double[T][N];
            digamma = new double[T][N][N];
            notes = Atom.toInt(args);
            alphaSum = 0;

            for (int i=0; i<T; ++i) {;
                System.out.println("Note " + i + " is " + notes[i] + ". ");
            }
            
            //compute alpha[0][i]
            //System.out.println(" ");
            //System.out.println("These are UNscaled alpha[0][i] probs, alpha[0][0], alpha[0][1]... Remember 0 refers to time 0...");
            c[0] = 0;
            for (int i=0; i<N; ++i) {
                alpha[0][i] = pi[i]*b[i][notes[0]];
                c[0] = c[0]+alpha[0][i];
                //System.out.println("alpha[0][" + i + "] is " + alpha[0][i]);
            }
        
            //scale the alpha[0][i]
            //System.out.println(" ");
            //System.out.println("These are scaled alpha[0][i] values, alpha[0][0], alpha[0][1]... Remember 0 refers to time 0...");
            c[0] = 1/c[0];
            for (int i=0; i<N; ++i) {
                alpha[0][i] = c[0] * alpha[0][i];
                //System.out.println("alpha[0][" + i + "] is " + alpha[0][i]);
            }
        
            //compute alpha[t][i]
            //System.out.println(" ");
            //System.out.println("These are scaled alpha[t][i] values, alpha[1][0], alpha[1][1]...");
            for (int t=1; t<T; ++t) {
                c[t] = 0;
                for (int i=0; i<N; ++i) {
                    alpha[t][i] = 0;
                    for (int j=0; j<N; ++j) {
                        alpha[t][i] = alpha[t][i]+alpha[t-1][j]*a[j][i];    //is this a[j][i] in the right order?? Check it!
                    }
                    alpha[t][i] = alpha[t][i]*b[i][notes[t]];   //Again needs to be an observation
                    c[t] = c[t]+alpha[t][i];
                    //System.out.println("alpha[" + t + "][" + i + "] is " + alpha[t][i]);  //these are the unscaled values of alpha
                }
                
                //sum the unscaled alpha[T-1][i]
                for (int i=0; i<N; ++i) {
                    alphaSum = alphaSum + alpha[T-1][i];
                }
            
                //scale alpha[t][i]
                c[t] = 1/c[t];
                for (int i=0; i<N; ++i) {
                    alpha[t][i] = c[t]*alpha[t][i];;
                    //System.out.println("alpha[" + t + "][" + i + "] is " + alpha[t][i]);    //these are the scaled values of alpha
                }
            }
        
            //Let beta[T-1][i]=1 scaled by c[T-1]
            //System.out.println(" ");
            //System.out.println("These scaled beta[T-1][i] = c[T-1]");
            for (int i=0; i<N; ++i) {
                //beta[T-1][i]=1; //This is used to see beta probs instead of unscaled values
                beta[T-1][i]=c[T-1];
                //System.out.println("beta[" + (T-1) + "][" + i + "] is " + beta[T-1][i]);
            }
        
            //Beta-pass
            //System.out.println(" ");
            //System.out.println("These are scaled beta[t][i] values, beta[T-2][0], beta[T-2][1]...");
            for (int t=T-2; t>=0; --t) {
                for (int i=0; i<N; ++i) {
                    beta[t][i] = 0;
                    for (int j=0; j<N; ++j) {
                        beta[t][i] = beta[t][i]+a[i][j]*b[j][notes[t+1]]*beta[t+1][j];
                    }
                    //scale beta[t][i] with the same scale factor as alpha[t][i]
                    beta[t][i] = c[t]*beta[t][i]; //Comment this out if want to see unscaled values
                    //System.out.println("beta[" + t + "][" + i + "] is " + beta[t][i]);
                }
            }
        
            //Compute digamma[t][i][j] and gamma[t][i] note, no scaling is done in these digamma
            //all scaling is done in the alpha and beta passes.
            //System.out.println(" ");
            //System.out.println("These are the values for digamma[t][i][j] and gamma[t][i]");
            for (int t=0; t<T-1; ++t) {
                denom = 0;
                for (int i=0; i<N; ++i) {
                    for (int j=0; j<N; ++j) {
                        denom = denom + alpha[t][i]*a[i][j]*b[j][notes[t+1]]*beta[t+1][j];
                    }
                }
            
                for (int i=0; i<N; ++i) {
                    gamma[t][i] = 0;
                    for (int j=0; j<N; ++j) {
                        digamma[t][i][j] = (alpha[t][i]*a[i][j]*b[j][notes[t+1]]*beta[t+1][j])/denom;
                        gamma[t][i] = gamma[t][i] + digamma[t][i][j];
                        //System.out.println("digamma[" + t + "][" + i + "][" + j +"] is " + digamma[t][i][j]);
                    }
                    //System.out.println("gamma[" + t + "][" + i + "] is " + gamma[t][i]);
                    //System.out.println(" ");
                }
                probOfSeq2 = denom;
            }
            
            logProb = 0;
            for (int t=0; t<T; ++t) {
                logProb = logProb + Math.log(c[t]);
            }
            logProb = -logProb;
            probOfSeq = Math.exp(logProb);
            
            System.out.println(" ");
            //System.out.println("A list in the right inlet");
            //System.out.println(" ");
            //System.out.println("According to the alpha's, P(O|Model) = " + alphaSum);
            //System.out.println("According to the denominators's, remember this is a scaled value, P(O|Model) = " + probOfSeq2);
            System.out.println("According to logProb, P(O|Model) is " + probOfSeq);
            System.out.println(" ");
            outlet(1, 1000000*probOfSeq);
        }

    }
}